import requests


def request(url, header=None, params=None):
    if header is None and params is None:
        return requests.get(url)
    elif header is None:
        return requests.get(url, params=params)
    elif params is None:
        return requests.get(url, headers=header)
    else:
        return requests.get(url, headers=header, params=params)


def request_json(url, header=None, params=None):
    response = request(url, header, params)
    return response.json()



import datetime
import time
from firstimpression.xml import get_text_from_element


def get_datetime_object_from_element(element, name, format):
    """ Gets the datetime object of an element (xml) based on name and format

    :param element: the element from which the date needs to be grabbed
    :type element: xml.etree.ElementTree.Element

    :param name: name of the element in xml which contains date
    :type name: str

    :param format: the format of the datetime in the element
    :type format: str

    :returns: time and date in the right format
    :rtype: date
     """
    unparsed_pubdate = get_text_from_element(element, name)

    created_at = parse_string_to_time(element, format)
    return parse_timestamp_to_date(created_at)


def parse_string_to_time(element, format):
    """ Gets the timestamp of an string in specific format

    :param element: string with the time in it
    :type element: str

    :param format: string with the format devined
    :type format: str

    :returns: time as timestamp
    :rtype: float """
    return parse_date_to_time(parse_string_to_date(element, format))


def parse_date_to_time(date_object):
    return time.mktime(date_object.timetuple())


def parse_string_to_date(element, format):
    return datetime.datetime.strptime(element, format)


def parse_timestamp_to_date(timestamp):
    """ Change timestamp to date object

    :param timestamp: the time in POSIX timestamp
    :type timestamp: float

    :returns: the time as a date
    :rtype: date """
    return datetime.datetime.fromtimestamp(timestamp)


def parse_date_to_string(date_object, format):
    return datetime.datetime.strftime(date_object, format)


def parse_string_time_to_minutes(element):
    [hours, minutes, seconds] = element.split(':')

    minutes = int(minutes)

    if int(seconds) > 30:
        minutes += 1

    minutes += int(hours) * 60

    return minutes


def get_month_text(month_number, month_index):
    return{
        1: ['jan', 'januari'],
        2: ['feb', 'februari'],
        3: ['maa', 'maart'],
        4: ['apr', 'april'],
        5: ['mei', 'mei'],
        6: ['jun', 'juni'],
        7: ['jul', 'juli'],
        8: ['aug', 'augustus'],
        9: ['sep', 'september'],
        10: ['okt', 'oktober'],
        11: ['nov', 'november'],
        12: ['dec', 'december']
    }[month_number][month_index]


def parse_date_to_string_full_day_month(date_object, month_type):
    day = str(date_object.day)
    month = get_month_text(date_object.month, month_type)
    return ' '.join([day, month, parse_date_to_string(date_object, '%Y %H:%M')])



import os
import time
import imghdr
import shutil
import requests
from firstimpression.scala import install_content
import xml.etree.ElementTree as ET


def create_directories(directories):
    for dirpath in directories:
        if not os.path.exists(dirpath):
            os.makedirs(dirpath)


def get_age(filepath):
    return time.time() - os.path.getmtime(filepath)


def check_too_old(filepath, max_age):
    try:
        file_age = get_age(filepath)
    except WindowsError:
        return True

    if file_age > max_age:
        return True
    else:
        return False


def check_valid_jpeg(filepath):
    return imghdr.what(filepath) == 'jpeg' or imghdr.what(filepath) == 'jpg'


def download_media(media_link, subdirectory, temp_folder):
    # Downloads and returns path of media
    media_filename = media_link.split('/').pop()
    media_path = os.path.join(temp_folder, subdirectory, media_filename)

    if not os.path.exists(media_path):
        response = requests.get(media_link, stream=True)
        with open(media_path, 'wb') as writefile:
            shutil.copyfileobj(response.raw, writefile)
    return media_path


def purge_directories(directories, max_days):
    # Remove all files from directory that are older than max_days
    for directory in directories:
        files = os.listdir(directory)
        for file in files:

            filepath = os.path.join(directory, file)
            file_age = get_age(filepath)

            if file_age > max_days * 86400:
                os.remove(filepath)


def write_root_to_xml_files(root, path, subfolder=None):
    tree = ET.ElementTree(root)
    tree.write(path)
    install_content(path, subfolder)


def xml_to_root(path):
    tree = ET.parse(path)
    root = tree.getroot()
    return root




def get_text_from_element(element, tag_name):
    return element.find(tag_name).text


def get_items_from_element(element, tag_name):
    return element.findall(tag_name)


def get_attrib_from_element(element, tag_name, attrib_name):
    return element.find(tag_name).attrib[attrib_name]



import datetime


def remove_tags_from_string(tags, text):
    return tags.sub('', text)


def replace_html_entities(replace_dict, text):
    for key in replace_dict:
        text = text.replace(key, replace_dict[key])
        text = text.strip()
    return text



import os

from firstimpression.file import download_media
from firstimpression.scala import install_content
from firstimpression.xml import get_attrib_from_element
from firstimpression.constants import PICTURE_FORMATS


def install_picture_content_wrap(picture_format, subdirectory, temp_folder, element, tag, attrib):
    # Installs content to LocalIntegratedContent folder and returns mediapath
    if not picture_format in PICTURE_FORMATS:
        return None

    media_link = get_attrib_from_element(
        element, tag, attrib).replace("_sqr256", "")

    if picture_format == 'square':
        media_link = media_link.replace('.jpg', '_sqr512.jpg')

    media_path = download_media(
        PICTURE_FORMATS[picture_format] + media_link, subdirectory, temp_folder)

    install_content(media_path, subdirectory)

    media_filename = media_path.split('\\').pop()

    return os.path.join('Content:\\', subdirectory, media_filename)



from geopy import distance
import xml.etree.ElementTree as ET

METER_TO_KILOMETER_DIVIDER = 1000
KILOMETER_APPENDIX = ' km'
UNKNOWN_JAM_DISTANCE_MESSAGE = 'onbekend'


def get_road_type(circumstance):
    try:
        return circumstance['roadNumber'][0]
    except KeyError:
        return False


def get_road(circumstance):
    try:
        return circumstance['roadNumber']
    except KeyError:
        return False


def get_road_number(circumstance):
    # Get road number (without A or N)
    try:
        return circumstance['roadNumber'][1:]
    except KeyError:
        return False


def get_from(circumstance):
    from_location = circumstance['directionText'].split(' - ')[0]
    return from_location


def get_to(circumstance):
    to_location = circumstance['directionText'].split(' - ')[1]
    return to_location


def get_length(circumstance):
    total_length = circumstance.get('total_length', None)
    if total_length is None:
        return ''
    else:
        return int(total_length)


def get_length_string(circumstance):
    length = circumstance.get('total_length', '')
    if length:
        return ''.join([str(int(length / METER_TO_KILOMETER_DIVIDER)), KILOMETER_APPENDIX])
    else:
        return UNKNOWN_JAM_DISTANCE_MESSAGE


def get_reason(circumstance):
    return circumstance.get('cause', '')


def get_area_detail(circumstance):
    return circumstance.get('locationText', '')


def get_event(circumstance):
    return circumstance.get('title', '')


def get_description(circumstance):
    return circumstance.get('description', '')


def get_type(circumstance):
    return circumstance.get('obstructionType', 0)


def get_total_delay_string(circumstance):
    # returns string with delay in minutes or False

    total_delay = circumstance.get('delay', '')

    if total_delay:
        # Round up minutes
        return '+{} min'.format(total_delay)
    else:
        return ''


def get_coordinates(circumstance):
    try:
        return {'longitude': circumstance['longitude'],
                'latitude': circumstance['latitude']
                }
    except KeyError:
        return False


def get_distance_to_circumstance(from_coordinates, to_coordinates):
    # Calculates distance from one coordinate to another (WATCH OUT: straight line, so no roads taken into account)
    if not from_coordinates or not to_coordinates:
        return ''

    coords_1 = (from_coordinates['latitude'], from_coordinates['longitude'])
    coords_2 = (to_coordinates['latitude'], to_coordinates['longitude'])

    return distance.distance(coords_1, coords_2).km


def parse_circumstance(circumstance, own_coordinates):
    # Parses JSON from API to own format. Junk data is removed.
    parsed_circumstance = dict()

    parsed_circumstance['road_type'] = get_road_type(circumstance)
    parsed_circumstance['road'] = get_road(circumstance)
    parsed_circumstance['road_number'] = get_road_number(circumstance)
    parsed_circumstance['from'] = get_from(circumstance)
    parsed_circumstance['to'] = get_to(circumstance)
    parsed_circumstance['length'] = get_length(circumstance)
    parsed_circumstance['length_string'] = ''
    parsed_circumstance['reason'] = get_reason(circumstance)
    parsed_circumstance['area_detail'] = get_area_detail(circumstance)
    parsed_circumstance['event'] = get_event(circumstance)
    parsed_circumstance['description'] = get_description(circumstance)
    parsed_circumstance['type'] = get_type(circumstance)
    parsed_circumstance['total_delay'] = get_total_delay_string(circumstance)
    parsed_circumstance['coordinates'] = get_coordinates(circumstance)

    parsed_circumstance['distance_to_circumstance'] = get_distance_to_circumstance(
        own_coordinates, parsed_circumstance['coordinates'])

    return parsed_circumstance


def parse_circumstances(circumstances, own_coordinates):
    parsed_circumstances = list()
    for circumstance in circumstances:
        parsed_circumstances.append(
            parse_circumstance(circumstance, own_coordinates))
    return parsed_circumstances


def parse_circumstances_to_XML(circumstances, exclude_items):
    # Parses list of json objects to XML
    root = ET.Element("root")
    for circumstance in circumstances:
        item = ET.SubElement(root, 'item')
        for attribute in circumstance:
            if attribute not in exclude_items:
                ET.SubElement(item, attribute).text = str(
                    circumstance[attribute])

    return root


def sort_longest_jams(jams):
    # Sorts longest jams starting with longest to shortest
    return sorted(jams, key=lambda i: i['length'], reverse=True)


def sort_closest_jams(jams):
    # Sorts jams that are closest to own location
    return sorted(jams, key=lambda i: i['distance_to_circumstance'])


def get_jams(circumstances, jam_type):
    # Filters jams from all possible circumstances which could be f.e. road works and/or police controls
    jams = list()
    for circumstance in circumstances:
        if circumstance['type'] == jam_type:
            jams.append(circumstance)
    return jams


def get_total_jam_length(circumstances):
    return circumstances['totalLengthOfJams'] / METER_TO_KILOMETER_DIVIDER


def get_total_jams_delay(jams):
    total_delay = 0
    for jam in jams:
        try:
            total_delay += int(jam['total_delay'].split(' ').pop(0))
        except (TypeError, ValueError):
            continue
    if total_delay > 60:
        return '{}+ uur'.format(int(total_delay / 60))
    else:
        return '{} min'.format(int(total_delay))


def get_total_length_string(total_length):
    return '{}{}'.format(total_length, KILOMETER_APPENDIX)


def get_only_highways(circumstances):
    specific_circumstances = list()
    for circumstance in circumstances:
        if circumstance['road_type'] == 'A':
            specific_circumstances.append(circumstance)
    return specific_circumstances



from tempfile import gettempdir

PICTURE_FORMATS = {
    'fullscreen': "http://narrowcasting.fimpweb.nl/imageresize.php?width=1920&url=",
    'square':     "http://narrowcasting.fimpweb.nl/imageresize.php?width=512&url=",
}

TEMP_FOLDER = gettempdir()
LOCAL_INTEGRATED_FOLDER = 'C:/Users/Public/Documents/Scala/LocalIntegratedContent'

APIS = ['', 'BBC','NS','NU','TRAFFIC']



from firstimpression import api
from firstimpression import file
from firstimpression import rss
from firstimpression import scala
from firstimpression import text
from firstimpression import time
from firstimpression import xml

from firstimpression import constants

from os.path import join

file.create_directories([join(constants.LOCAL_INTEGRATED_FOLDER, api_name) for api_name in constants.APIS])



from firstimpression.api.request import request_json
from firstimpression.time import parse_string_to_date, parse_string_time_to_minutes, parse_date_to_string


def get_response(url, headers, params):
    response_json = request_json(url, headers, params)

    if response_json.get('statusCode', None) == 429:
        raise Exception('Rate limit exceeded')

    return response_json


def get_departures(response_json):
    return response_json['payload']['departures']


def get_departure_time(departure):
    return departure['plannedDateTime'][:-5]


def get_actual_departure_time(departure):
    return departure['actualDateTime'][:-5]


def get_departure_number(departure):
    return departure['product']['number']


def get_destination(departure):
    return departure['direction']


def get_train_category(departure):
    return departure['product']['longCategoryName']


def get_route_text(departure):
    # Returns string with stations on route in this format: '{station}, {station}, {station}'
    return ', '.join([station['mediumName'] for station in departure['routeStations']])


def get_operator(departure):
    return departure['product']['operatorName']


def get_planned_track(departure):
    return departure['plannedTrack']


def get_actual_track(departure):
    return departure.get('actualTrack', '')


def get_delay(departure, date_format):
    try:
        if departure['cancelled'] == True:
            return 'Rijdt niet'
    except KeyError:
        pass

    planned_departure_time = parse_string_to_date(
        get_departure_time(departure), date_format)
    actual_departure_time = parse_string_to_date(
        get_actual_departure_time(departure), date_format)

    if planned_departure_time < actual_departure_time:
        delayed_time = actual_departure_time - planned_departure_time
        delayed_minutes = parse_string_time_to_minutes(str(delayed_time))
        return ''.join(['+', str(delayed_minutes), ' min'])
    else:
        return ''


def get_message(departure):
    try:
        message = departure.get('messages', False)
        if message:
            msg = message[0]['message']
        else:
            msg = ''
    except KeyError:
        msg = ''
    return msg


def get_parsed_departures(response_json, date_format):
    departures = get_departures(response_json)
    parsed_departures = list()
    for departure in departures:
        parsed_departure = dict()
        parsed_departure['departure_number'] = get_departure_number(departure)
        parsed_departure['departure_time'] = get_departure_time(departure)
        parsed_departure['destination'] = get_destination(departure)
        parsed_departure['train_category'] = get_train_category(departure)
        parsed_departure['route_text'] = get_route_text(departure)
        parsed_departure['operator'] = get_operator(departure)
        parsed_departure['planned_track'] = get_planned_track(departure)
        parsed_departure['actual_track'] = get_actual_track(departure)
        parsed_departure['delay'] = get_delay(departure, date_format)
        parsed_departure['message'] = get_message(departure)
        parsed_departures.append(parsed_departure)

    return parsed_departures



from setuptools import find_packages, setup
from os import path

HERE = path.abspath(path.dirname(__file__))

with open(path.join(HERE, 'README.md')) as f:
    README = f.read()

setup(
    name='firstimpression',
    packages=find_packages(),
    version='0.5.5',
    description='First Python library',
    long_description=README,
    long_description_content_type='text/markdown',
    author='FirstImpression',
    author_email='programming@firstimpression.nl',
    license='MIT',
    install_requires=["requests", "geopy", "wheel"],
    include_package_data=True,
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Build Tools',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 2.7'
    ],
)



"""scalascript.py - (C) Scala

A module to provide Python 3 compatibility with the Scala Linux Player's
embedded Python module.
"""
from copy import deepcopy
try:
    from collections.abc import MutableMapping, MutableSequence
except ImportError:
    from collections import MutableMapping, MutableSequence
import win32api
import win32com.client
import pythoncom
import logging
import traceback
import struct
import os

if not os.name == 'nt':
    raise ImportError('This module is intended for Microsoft Windows only.')

import sys
import inspect

# Don't allow importing from the command line or interactive interpreter
# as this module simulates the Linux Player's built-in scalascript
# module and shouldn't be available.

# Scala COM handles, initialized on demand
_scalaplayer = None
_infoplayer = None

# ActiveX ProgIDs
_prog_id_scalaplayer = 'ScalaPlayer.ScalaPlayer.1'
_prog_id_infoplayer5 = 'Scala.InfoPlayer5'
_prog_id_filelock = 'ScalaFileLock.ScalaFileLock.1'
_prog_id_netic = 'Scala.InfoNetwork'


def _hresult_as_hex(hr):
    # Courtesy of: https://stackoverflow.com/a/10641263/100955
    return struct.unpack("L", struct.pack("l", hr))[0]


class variables(MutableMapping):
    """This class matches the API baked-in on the Linux Player. It 
    behaves like a dictionary. Unlike the Python 2 implementation, this
    class provides access to ALL non-hidden ScalaScript variables, and 
    with their original names.

    Behavior that differs from sharedvars:
        - Does not handle default values
        - Raises KeyError when getting variables that don't exist
        - Only the shared variables are available in this dictionary,
          and NOT all global namespace items

    Behavior that differs from Linux player:
        - Any Explicitly shares Scala vars continue to be available in 
          the main namespace. Watch out for clobbering them!

    Example::

        import scalascript
        svars = scalalib.variables()
        f = file(svars['filename'])       # get value
        svars['result'] = 15              # set value

    """

    class array(MutableSequence):
        """Needs to behave like other mutable types."""

        def __init__(self, key):
            self._key = key

            # Start up the COM interface
            global _scalaplayer
            if not _scalaplayer:
                _scalaplayer = win32com.client.Dispatch(_prog_id_scalaplayer)
            self._scalaplayer = _scalaplayer

        def __repr__(self):
            item = None
            try:
                item = self._scalaplayer.GetScalaScriptValue(self._key)
            except Exception as err:
                log(logging.ERROR, str(err))

            return list(item)

        def __str__(self):
            return str(self.__repr__())

        def __getitem__(self, index):
            item = self.__repr__()
            return item[index]

        def __setitem__(self, index, value):
            log(logging.DEBUG, "Setting %s[%s] to %s", self._key, index, value)
            item = self.__repr__()
            item[index] = value
            try:
                self._scalaplayer.SetScalaScriptValue(self._key, tuple(item))
            except Exception as err:
                log(logging.ERROR, str(err))

        def __delitem__(self, index):
            raise IndexError("deleting ScalaScript array items is not allowed")

        def __len__(self):
            return len(self.__repr__())

        def insert(self, index, value):
            raise IndexError(
                "inserting ScalaScript array items is not allowed")

        def clear(self):
            raise IndexError("deleting ScalaScript array items is not allowed")

        def extend(self, values):
            raise IndexError(
                "extending ScalaScript array items is not allowed")

        def pop(self, index=-1):
            raise IndexError("deleting ScalaScript array items is not allowed")

        def remove(self, value):
            raise IndexError("removing ScalaScript array items is not allowed")

    def __init__(self, defaults=None):
        # Start up the COM interface
        global _scalaplayer
        if not _scalaplayer:
            _scalaplayer = win32com.client.Dispatch(_prog_id_scalaplayer)
        self._scalaplayer = _scalaplayer

        # Copy the defaults dict
        if defaults:
            self._defaults = deepcopy(defaults)
        else:
            self._defaults = {}

    def __delitem__(self, key):
        """Can't delete, move along."""
        pass

    def __iter__(self):
        for key in self._scalaplayer.ListScalaScriptValues():
            yield key

    def keys(self):
        return self._scalaplayer.ListScalaScriptValues()

    def __len__(self):
        return len(self._scalaplayer.ListScalaScriptValues())

    def __getitem__(self, key):
        item = None

        if key in self._scalaplayer.ListScalaScriptValues():
            try:
                item = self._scalaplayer.GetScalaScriptValue(key)
                if type(item) in (tuple, list):
                    item = self.array(key)
            except Exception as err:
                log(logging.ERROR, str(err))
        else:
            # Look in the defaults dict
            # Allow any KeyError here to be raised
            item = self._defaults[key]

        return item

    def __setitem__(self, key, item):
        """Set new values for existing items only.  Drop any newly added 
        values.
        """
        if self[key] is None:
            return

        result = None
        try:
            if type(item) in (tuple, list):
                # Create a scalascript.variables.array and set the new
                # values iteratively
                _array = self.array(key)
                for i, element in enumerate(item):
                    _array[i] = element
            else:
                # Assign the value
                result = self._scalaplayer.SetScalaScriptValue(key, item)
        except Exception as err:
            log(logging.ERROR, str(err))

    def items(self):
        for key in self.keys():
            yield key, self.__getitem__(key)

    def __contains__(self, key):
        return (key in self._scalaplayer.ListScalaScriptValues())


def find_content(scalapath):
    r"""Find platform path to versionated files via Scala-style virtual path.

    Note: 
        Windows and Linux require differently-formatted Scala paths.
        This function will convert appropriately, while the Linux player
        WILL NOT:

            # Works on Windows and Linux
            local_path = scalascript.getfilepath("Content://findme.txt")

            # Will not work on Linux
            local_path = scalascript.getfilepath("Content:\\findme.txt")

        Be mindful of case-sensitivity issues.

        Under the hood, this function calls LockScalaFile, and 
        UnlockScalaFile when the returned object goes out of scope.

    Args:
        scalapath (str): Scala-formatted path

    Returns:
        str: local filesystem path to versionated file.
    """
    # PyErr_SetString(PyExc_TypeError, "path must be a string");
    if not isinstance(scalapath, str):
        raise TypeError("path must be a string")

    _path = scalapath.replace("//", "\\")
    platform_path = _lock_content(_path)
    if platform_path:
        return platform_path
    else:
        raise FileNotFoundError(
            'The path "{}" does not exist.'.format(scalapath))


def install_content(abspath, subfolder=None):
    r'''Copies a file to the local Scala Content: folder.

    Note: Assumes netic service is already running!

    Args:
        abspath (str): A file (specified by its absolute path) to copy.
        subfolder (str): Place it into a subfolder to group related files.
    Notes:
        Installed files may be found later using a Scala-style virtual path
        string via find_content::

            find_content(r'Content://filename.ext')

    Returns:
        str: platform path to content, as installed and versionated.
    '''
    if not os.path.exists(abspath):
        raise FileNotFoundError('File "{}" does not exist.'.format(abspath))

    if not subfolder:
        subfolder = ''

    try:
        netic = win32com.client.Dispatch(_prog_id_netic)
        netic.IntegrateContentLocally(abspath, subfolder)
        log(logging.INFO, abspath)

        if subfolder:
            where_it_should_be = 'Content://' + \
                subfolder + '/' + os.path.basename(abspath)
        else:
            where_it_should_be = 'Content://' + os.path.basename(abspath)

        return find_content(where_it_should_be)

    except pythoncom.com_error as err:
        errstr = '{}\n\nPlayer Transmission Client ({}) not available.'.format(
            str(err), _prog_id_netic)
        log(logging.error, errstr.replace('\n', '  '))
        raise RuntimeError(errstr)
    except Exception as err:
        log(logging.error, str(err))


def _lock_content(scalapath, report_err=True):
    r"""Locks a content file so the player doesn't try to remove it.

    Args:
        scalapath (str): Scala-style virtual path, such as "Content://file.ext"
        report_err (bool): Log errors?

    Returns:
        str: The Windows path to the affected file.
             If the file is not found or there is an error, returns None.
    """
    class _StringAndLock(str):
        """A subclass of a string that carries a Scala lock object around with
        it.  The lock will be unlocked upon deletion when the object falls out of
        scope or is deleted explicitly.
        """

        def __del__(self):
            if hasattr(self, 'lockObj'):
                log(logging.DEBUG, 'unlock_content: "{}"'.format(scalapath))
                self.lockObj.UnlockScalaFile()

    try:
        lockObj = win32com.client.Dispatch(_prog_id_filelock)
        windows_path = lockObj.LockScalaFile(scalapath)

        # Add the lock into the string, to unlock upon its deletion.
        windows_path = _StringAndLock(windows_path)
        windows_path.lockObj = lockObj

        log(logging.INFO, '"{}" @ "{}"'.format(scalapath, windows_path))

        return windows_path
    except Exception as err:
        if report_err:
            log(logging.ERROR, str(err))
        return None


def log(level, msg, *args, **kwargs):
    """This logging method is meant to mimic the way the Linux Player 
    handles logging, and does not work the way Windows scalalib-based
    logging previously did. 

    Behavior that differs from Linux player:
        As the method made available via ActiveX does not allow for
        setting the loglevel, all messages are forced to "INFO".

        The Python2 scalalib module's logging.Handler inserted the
        loglevel and scope on an INFO line before the message, that
        behavior has been duplicated.

    Args:
        level (logging.loglevel): Somewhat ignored, but is inserted.
                                  See behvior notes above.
        msg (str): message to log in Python logging module format
        args: arguments for format string
        exc_info (bool, exception): Mimics logging module behavior.

    Note:
        This is not a full logging module object!
    """
    global _scalaplayer
    if not _scalaplayer:
        _scalaplayer = win32com.client.Dispatch(_prog_id_scalaplayer)

    try:
        attr_dict = {}
        attr_dict['level'] = level
        attr_dict['msg'] = msg
        attr_dict['args'] = args

        _exc_text = None
        exc_info = kwargs.get('exc_info')
        if exc_info:
            if isinstance(exc_info, BaseException):
                exc_info = (type(exc_info), exc_info, exc_info.__traceback__)
            elif not isinstance(exc_info, tuple):
                exc_info = sys.exc_info()

            attr_dict['exc_info'] = exc_info
            _exc_text = traceback.format_exception(*exc_info)

        _formatted_msg = logging.makeLogRecord(attr_dict).getMessage()
        if _exc_text:
            if _formatted_msg[-1:] != "\n":
                _formatted_msg += "\n"
            _formatted_msg += ''.join(_exc_text).strip()

        _caller = inspect.currentframe().f_back.f_code.co_name
        _level_name = logging.getLevelName(level)
        _scalaplayer.Log("{} {}: {}".format(
            _level_name, _caller, _formatted_msg))

    except pythoncom.com_error as err:
        # Do not attempt to log while logging is in an error state!
        # Not sure where this print actually goes, though.
        print('COM Error: {}'.format(err))


def quitplayer():
    """Send Quit Request to Scala Player.

    Returns:
        bool: False on failure
    """
    global _infoplayer
    try:
        # Comment from scalalib: don't cache obj, doesn't work more than once
        _infoplayer = win32com.client.Dispatch(_prog_id_infoplayer5)
        log(logging.DEBUG, 'Attempting to quit...')
        _infoplayer.Quit()
        return True
    except pythoncom.com_error as err:
        errstr = '{}\n\nScala Player ({}) not available.'.format(
            str(err), _prog_id_infoplayer5)
        log(logging.ERROR, errstr.replace('\n', '  '))
    except Exception as err:
        log(logging.ERROR, err)

    return False


def restartplayback():
    """Send Restart Playback request to Scala Player.

    Returns:
        bool: False on failure
    """
    global _infoplayer
    try:
        # Comment from scalalib: don't cache obj, doesn't work more than once
        _infoplayer = win32com.client.Dispatch(_prog_id_infoplayer5)
        log(logging.DEBUG, 'Attempting to restart playback...')
        _infoplayer.RestartPlayback()
        return True
    except pythoncom.com_error as err:
        errstr = '{}\n\nScala Player ({}) not available.'.format(
            str(err), _prog_id_infoplayer5)
        log(logging.ERROR, errstr.replace('\n', '  '))
    except Exception as err:
        log(logging.ERROR, err)

    return False


def sleep(secs):
    """Sleep lightly for a number of seconds while maintaining awareness of 
    whether or not playback has been aborted.

    Note:
        Follows the Python convention and takes SECONDS, not milliseconds. 
        This makes it easier to conditionally import scalascript.sleep or 
        time.sleep for command-line testing.

    Raises:
        KeyboardInterrupt

    Args:
        secs (float): fractional seconds
    """
    DISP_E_EXCEPTION = 0x80020009
    E_ABORT = 0x80004004

    global _scalaplayer
    if not _scalaplayer:
        _scalaplayer = win32com.client.Dispatch(_prog_id_scalaplayer)

    try:
        _scalaplayer.Sleep(secs * 1000)
    except pythoncom.com_error as err:
        # On abort, raise KeyboardInterrupt, like the Linux Player does
        # when it receives SIGINT.
        # scalalib.sleep() would call sys.exit(9)

        # Need to check main hresult and the exception info, as the main hresult
        # may just be an iDispatch general error.
        if _hresult_as_hex(err.hresult) == E_ABORT:
            raise KeyboardInterrupt
        elif _hresult_as_hex(err.hresult) == DISP_E_EXCEPTION and _hresult_as_hex(err.excepinfo[-1]) == E_ABORT:
            raise KeyboardInterrupt
        else:
            raise
